#!/usr/bin/env python

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid

matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype']  = 42
matplotlib.rcParams['font.family'] = "serif"
matplotlib.rcParams['font.size'] = 10

from square import square
from Multinomial import Multinomial

import numpy as np
import sys

from itertools import permutations

def artificial_dataset( N, k, seed, dim, sigma, eps=1e-16 ):
    '''
        N -- sample size
        k -- number of clusters
     seed -- random seed
      dim -- dimension of the probability vector
    sigma -- noise level

    return a list of Multinomail distributions and their labels
    '''
    assert( k > 0 )
    np.random.seed( seed )

    theta = []
    labels = []

    cluster_size = [ int(N/k) for _ in range(k-1) ]
    cluster_size.append( N - sum(cluster_size) )

    for i, _n in enumerate( cluster_size ):
        center = np.random.dirichlet( np.ones(dim) )
        theta.append( sigma * np.random.randn( _n, dim ) + np.log( center+eps ) )
        labels += [ i for _ in range(_n) ]

    theta = np.vstack( theta )
    eta = np.exp( theta )
    return [ Multinomial(p) for p in eta ], labels

def visualize_clusters( N=100, num_clusters=3, RES=300, dim=3, sigma=0.3, seed=2001 ):
    '''
    show toy clustering results
    '''
    distributions, labels = artificial_dataset( N, num_clusters, seed, dim, sigma )

    fig = plt.figure( figsize=(10,3), dpi=300 )
    grid = ImageGrid( fig, 111,
                 nrows_ncols=(1,3),
                 axes_pad=0.5,
                 share_all=True,
                 cbar_location="right",
                 cbar_mode="each",
                 cbar_size="3%",
                 cbar_pad=0,
                 )

    def showdist( ax, clustering, compute_center, compute_distance, title ):
        assign = clustering( distributions, num_clusters, seed=seed )
        centers = []
        for cluster_idx in range( num_clusters ):
            cluster = [ distributions[i] for i,idx in enumerate(assign) if np.isclose(idx, cluster_idx) ]
            centers.append( compute_center( cluster ) )

        _distance = np.array( [ square( c, compute_distance, RES ) for c in centers ] ).min(0)

        vmax = _distance[np.logical_not(np.isinf(_distance))].max() * .5
        im   = ax.imshow( _distance, origin='lower', cmap='afmhot_r', vmin=0, vmax=vmax )
        cbar = ax.cax.colorbar( im )

        levels = [ 0.2*i for i in range( 1, 20 ) ]
        ax.contour( _distance, origin='lower', cmap='afmhot_r', levels=levels, linewidths=1, alpha=.5 )

        for cluster_idx in range( num_clusters ):
            cluster = [ distributions[i] for i,idx in enumerate(assign) if np.isclose(idx, cluster_idx) ]
            c = compute_center( cluster )
            enclosing = max( [ compute_distance( _d, c ) for _d in cluster ] )
            ax.contour( square( c, compute_distance, RES ), origin='lower', cmap='YlGn_r', levels=[enclosing], linewidths=2, alpha=.5 )

        markers = [ 'o', '1', '*' ]
        colors  = [ 'r', 'b', 'w' ]
        for _, _dist in zip( assign, distributions ):
            eta = _dist.eta()
            ax.scatter( eta[0]*RES, eta[1]*RES, color=colors[_], edgecolor='k', linewidths=.5, marker=markers[_], s=16, alpha=.8 )

        ax.set_xticks( [0, RES] )
        ax.set_xticklabels( ['0','1'] )
        ax.set_yticks( [0, RES] )
        ax.set_yticklabels( ['0','1'] )
        ax.set_title( title )

    showdist( grid[0], Multinomial.riemannian_kcenters, Multinomial.riemannian_center, Multinomial.riemannian_distance, r'$\rho_{\mathrm{FHR}}$' )
    showdist( grid[1], Multinomial.kl_kcenters, Multinomial.kl_center, Multinomial.kl_distance, r'$\rho_{\mathrm{IG}}$' )
    showdist( grid[2], Multinomial.hilbert_kcenters, Multinomial.hilbert_center, Multinomial.hilbert_distance, r'$\rho_{\mathrm{HG}}$' )
    fig.savefig( 'clustering.pdf', bbox_inches='tight' )

def accuracy( assign, truth, k ):
    '''
    measure the clustering accuracy
    '''
    assert( len(assign) == len(truth) )

    max_acc = 0
    for per in permutations( range( k ) ):
        acc = 0
        for a, b in zip( assign, truth ):
            if per[a] == b:
                acc += 1
        if acc > max_acc: max_acc = acc

    return( max_acc/len(assign) )

def exp( alg, dim, sigma, N, num_clusters, max_center_itrs=50, repeat=50 ):
    '''
             alg -- algorithm
             dim -- simplex dimension
           sigma -- noise level
               N -- sample size
    num_clusters -- number of clusters
    max_center_itrs -- number of iterations during center computation
          repeat -- number of repeats
    '''
    algonames = [ 'Riemannian', 'KL', 'KL(kmeans)', 'Hilbert', 'Euclidean' ]

    print('============================')
    print( 'running {0} for {1} times'.format( algonames[alg], repeat ) )
    print( '#samples={0}'.format(N) )
    print( '#dim={0}'.format(dim) )
    print( 'sigma={0}'.format(sigma) )
    print( '#clusters={0}'.format(num_clusters) )

    acc = np.zeros( repeat, dtype=np.float )
    for run in range( repeat ):
        magic_number = 2017+run
        distributions, labels = artificial_dataset( N, num_clusters, magic_number, dim, sigma )

        if alg == 0:
            assign = Multinomial.riemannian_kcenters( distributions, num_clusters, seed=magic_number, max_center_itrs=max_center_itrs )
        elif alg == 1:
            assign = Multinomial.kl_kcenters( distributions, num_clusters, seed=magic_number, max_center_itrs=max_center_itrs )
        elif alg == 2:
            assign = Multinomial.kl_kmeans( distributions, num_clusters, seed=magic_number, max_center_itrs=max_center_itrs )
        elif alg == 3:
            assign = Multinomial.hilbert_kcenters( distributions, num_clusters, seed=magic_number, max_center_itrs=max_center_itrs )
        elif alg == 4:
            assign = Multinomial.euclidean_kcenters( distributions, num_clusters, seed=magic_number, max_center_itrs=max_center_itrs )
        else:
            raise RuntimeError( 'unknown algorithm' )
        acc[run] = accuracy( assign, labels, num_clusters )

    print( 'accuracy:', acc.mean(), acc.std() )
    print( '' )

if __name__ == '__main__':
    if len( sys.argv ) < 2:
        print( 'usage: %s [0-4V]' % sys.argv[0] )

    elif sys.argv[1].lower() == 'v':
        visualize_clusters()

    else:
        expid = int( sys.argv[1] )
        for num_clusters in [5,]:
            for N in [ 100,]:
                for dim in [ 10, 256 ]:
                    for sigma in [ 0.5, 0.7, 0.9, 1.1, 1.3, 1.5 ]:
                        exp( expid, dim, sigma, N, num_clusters )

